"""Deployment related datastructures"""
import typing as t

from pydantic import BaseModel

from taktile_types.enums.instance import InstanceClass, InstanceSize


class DeploymentMetadata(BaseModel):
    """
    Metadata of the deployment

    Attributes
    ----------

    name: str
        name used by kubernetes
    """

    name: str


class DeploymentSpec(BaseModel):
    """
    Full specification of the deployment

    Attributes
    ----------

    ref: str
        Branch or tag
    commit: str
        Full commit hash
    repository_uri: str
        URI of repository on github
    principal: str
        user or organization which own the repo
    rest_instance_size: Optional[InstanceSize]
        enumeration of size
    rest_instance_type: Optional[InstanceType]
        enumeration of type
    rest_cpu_request: Optional[str]
        kubernetes-parsable cpu request
    rest_memory_request: Optiona[str]
        kubernetes-parsable memory request
    arrow_instance_size: Optional[InstanceSize]
        enumeration of size
    arrow_instance_type: Optional[InstanceType]
        enumeration of type
    arrow_cpu_request: Optional[str]
        kubernetes-parsable cpu request
    arrow_memory_request: Optiona[str]
        kubernetes-parsable memory request
    rest_replicas: int
        min count for rest replicas
    max_rest_replicas: int
        max scale for rest replicas
    arrow_replicas: int
        static scale for arrow replicas

    repository_key: str
    repository_id: str
        uuid for associated repository

    github_auth_volume: str
        name of volume containing github token
    taktile_auth_volume: str
        name of volume containing taktile token
    api_key_auth_volume: str
        name of volume containing api key
    repository_secret_auth_volume: str
        name of volume containing repo secrets


    """

    ref: str
    commit: str
    repository_uri: str

    principal: str

    rest_instance_size: t.Optional[InstanceSize]
    rest_instance_type: t.Optional[InstanceClass]
    rest_cpu_request: t.Optional[str]
    rest_memory_request: t.Optional[str]
    rest_gpu_request: t.Optional[str]

    arrow_instance_size: t.Optional[InstanceSize]
    arrow_instance_type: t.Optional[InstanceClass]
    arrow_cpu_request: t.Optional[str]
    arrow_memory_request: t.Optional[str]
    arrow_gpu_request: t.Optional[str]

    rest_replicas: int
    max_rest_replicas: int
    arrow_replicas: int

    repository_key: str
    repository_id: str

    github_auth_volume: str
    taktile_auth_volume: str
    api_key_auth_volume: str
    repository_secret_auth_volume: str


class Deployment(BaseModel):
    """
    Top-level TD Object

    Attributes
    ----------

    apiVersion: str
        version of the CRD to use
    kind: str
        identifier for CRD type
    spec: DeploymentSpec
        full spec for TD
    metadata: DeploymentSpec
        object meta including name
    """

    apiVersion: str
    kind: str
    spec: DeploymentSpec
    metadata: DeploymentMetadata


class SecretMetadata(BaseModel):
    """
    Secret object metadata

    Attributes
    ----------
    name: str
        name of object
    """

    name: str


class RepositorySecret(BaseModel):
    """
    Secret contents

    Attributes
    ----------
    secret_name: str
        name of individual secret
    secret_value: str
        content of individual secret

    """

    secret_name: str
    secret_value: str


class Secret(BaseModel):
    """
    Top-level secret object

    Attributes
    ----------
    token: str
        secret token
    metadata: SecretMetadata
        identifier for object
    """

    token: str
    metadata: SecretMetadata


class ApplyPayload(BaseModel):
    """
    Fully serialization of deployment payload

    Attributes
    ----------
    secret: Secret
        secret used in update
    deployment: Deployment
        Full TD object
    repository_secret: List[RepositorySecret]
        list of all repo secrets
    update_key: str
        key to write back to taktile APIs
    api_key: str
        key to authenticate against model
    """

    secret: Secret
    deployment: Deployment
    repository_secrets: t.List[RepositorySecret]
    update_key: str
    api_key: str



"""Instance sizes and types."""
from .common import ExtendedEnum


class InstanceSize(ExtendedEnum):
    """The size values of instances."""

    SMALL = "small"
    MEDIUM = "medium"
    LARGE = "large"
    XLARGE = "xlarge"
    XXLARGE = "xxlarge"
    XXXLARGE = "xxxlarge"


class InstanceClass(ExtendedEnum):
    """Instance classes of Taktile Deployments.

    * GP: general purpose (deprecated)
    * GPU: deployment with a GPU attached
    * CPU: deployment with 1 cpu but varying RAM sizes
    """

    GP = "gp"
    GPU = "gpu"
    CPU = "cpu"


class InstanceType(ExtendedEnum):
    """Valid instance types for Taktile Deployments"""

    GP_SMALL = "gp.small"
    GP_MEDIUM = "gp.medium"
    GP_LARGE = "gp.large"
    GP_XLARGE = "gp.xlarge"
    GP_2XLARGE = "gp.xxlarge"
    GPU_SMALL = "gpu.small"
    CPU_SMALL = "cpu.small"
    CPU_MEDIUM = "cpu.medium"
    CPU_LARGE = "cpu.large"
    CPU_XLARGE = "cpu.xlarge"
    CPU_2XLARGE = "cpu.xxlarge"
    CPU_3XLARGE = "cpu.xxxlarge"


class ServiceType(ExtendedEnum):
    """Enum for service types."""

    REST = "rest"
    GRPC = "grpc"



"""Common enum functionality."""

import typing as t
from enum import Enum


class ExtendedEnum(str, Enum):
    """ExtendedEnum.

    Enums with an easy way of querying all members.
    """

    @classmethod
    def set(cls) -> t.Set[str]:
        """set.

        Parameters
        ----------

        Returns
        -------
        t.Set[str]
            A set of all values.

        """
        return {c.value for c in cls}

    @classmethod
    def list(cls) -> t.List[str]:
        """list.

        Parameters
        ----------

        Returns
        -------
        t.List[str]
            A list of all values.

        """
        return [c.value for c in cls]

    @classmethod
    def names(cls) -> t.List[str]:
        """names.

        Parameters
        ----------

        Returns
        -------
        t.List[str]
            A list of all enum names.

        """
        return [c.name for c in cls]



"""Enums about endpoints"""

from taktile_types.enums.common import ExtendedEnum


class EndpointKinds(ExtendedEnum):
    """EndpointKinds.
    Different kinds of endpoints taktile-components supports
    """

    GENERIC = "generic"
    TYPED = "typed"
    ARROW = "arrow"
    PROFILED = "profiled"


class ProfileKinds(ExtendedEnum):
    """ProfileKinds.
    Different profile kinds supported by taktile-profiling
    """

    BINARY = "binary"
    REGRESSION = "regression"


class ArrowFormatKinds(ExtendedEnum):
    """ArrowFormatKinds.
    Return types of an arrow / profiled endpoint
    """

    DATAFRAME = "dataframe"
    SERIES = "series"
    ARRAY = "array"



"""
Schemas related to model unit tests
"""
import typing as t

from pydantic import UUID4, BaseModel

from taktile_types.enums.deployment.test import TestOutcomeColor, TestTypes


class TestResultSpecBase(BaseModel):
    """TestResultSpecBase.
    Base model of a test suite

    Attributes
    ----------
    test_suite_id : UUID4
        unique identifier
    outcome : str
        outcome of this test suite. TODO: This should be properly typed.

    """

    test_suite_id: UUID4
    outcome: str


class TestResultSpec(TestResultSpecBase):
    """TestResultSpec.
    Extended model of a test suite

    Attributes
    ----------
    section_name : str
        section name in which this test is
    message: str
        message of this test (either failure or success message)
    file_path: t.Optional[str]
        file path of the test file this is from
    test_name: t.Optional[str]
        name of of the test
    test_type: TestTypes
        type of test. TODO: Remove this, we don't use it.
    """

    section_name: str
    message: str
    file_path: t.Optional[str]
    test_name: t.Optional[str]
    test_type: TestTypes


class TestResultResponse(BaseModel):
    """TestResultResponse.
    Deployment-Api response model for tests

    Attributes
    ----------
    color : TestOutcomeColor
        a simplified color response: TODO: Remove this, we don't use it.
    git_hash : str
        git hash of this model
    results : t.List[TestResultSpec]
        list of test results
    """

    color: TestOutcomeColor
    git_hash: str
    results: t.List[TestResultSpec]



"""Access enums for repositories"""
from taktile_types.enums.common import ExtendedEnum


class AccessKind(ExtendedEnum):
    """Kinds of access to a repository"""

    OWNER = "owner"
    VIEWER = "viewer"



"""Resource related datastructures"""
import re
import typing as t

from pydantic import BaseModel, validator

RESOURCE_REGEX = re.compile("^([+-]?[0-9.]+)([eEinumkKMGTP]*[-+]?[0-9]*)$")


class ResourceRequirement(BaseModel):
    """ResourceRequirement
    A data structure to hold the cpu, memory, and gpu requirements of a service
    """

    cpu: t.Optional[str] = None
    memory: t.Optional[str] = None
    gpu: str = "0"

    @validator("gpu")
    def gpu_validator(
        cls, val
    ):  # pylint: disable=no-self-use,no-self-argument
        """Validates gpu resource request"""
        num = int(val)
        if num < 0:
            raise ValueError("GPU request must be a positive integer")
        return val

    @validator("cpu")
    def cpu_validator(
        cls, val
    ):  # pylint: disable=no-self-use,no-self-argument
        """Validates cpu resource request"""
        if not RESOURCE_REGEX.match(val):
            raise ValueError(
                "CPU request must match regex:"
                " ^([+-]?[0-9.]+)([eEinumkKMGTP]*[-+]?[0-9]*)$"
            )
        return val

    @validator("memory")
    def memory_validator(
        cls, val
    ):  # pylint: disable=no-self-use,no-self-argument
        """Validates memory resource request"""
        if not RESOURCE_REGEX.match(val):
            raise ValueError(
                "Memory request must match regex:"
                " ^([+-]?[0-9.]+)([eEinumkKMGTP]*[-+]?[0-9]*)$"
            )
        return val


SIZE_MAP = {
    "gp": {
        "small": ResourceRequirement(cpu="500m", memory="512Mi", gpu="0"),
        "medium": ResourceRequirement(cpu="1000m", memory="1Gi", gpu="0"),
        "large": ResourceRequirement(cpu="2000m", memory="2Gi", gpu="0"),
        "xlarge": ResourceRequirement(cpu="2000m", memory="4Gi", gpu="0"),
        "xxlarge": ResourceRequirement(cpu="2000m", memory="8Gi", gpu="0"),
    },
    "cpu": {
        "small": ResourceRequirement(cpu="500m", memory="512Mi", gpu="0"),
        "medium": ResourceRequirement(cpu="1000m", memory="1Gi", gpu="0"),
        "large": ResourceRequirement(cpu="1000m", memory="2Gi", gpu="0"),
        "xlarge": ResourceRequirement(cpu="1000m", memory="3Gi", gpu="0"),
        "xxlarge": ResourceRequirement(cpu="1000m", memory="4Gi", gpu="0"),
        "xxxlarge": ResourceRequirement(cpu="1000m", memory="8Gi", gpu="0"),
    },
    "gpu": {
        "small": ResourceRequirement(cpu="3500m", memory="14Gi", gpu="1"),
    },
}



"""Request Scope related datastructures"""
import typing as t

from pydantic import BaseModel, Field, validator

from taktile_types.enums.http import Method


class Headers(BaseModel):
    """HTTP Request Header Information"""

    user_agent: str = Field("", alias="user-agent")

    @validator("user_agent")
    def trunc(cls, val):  # pylint: disable=no-self-use,no-self-argument
        """Trims lengthy browser user-agents"""
        if val:
            return val.split()[0]
        return val


class Scope(BaseModel):
    """HTTP Request Scope Information"""

    client: t.Tuple[str, str] = ("", "")
    method: Method = Method.GET
    path: str = ""
    type: str = ""
    http_version: str = ""

    @validator("type")
    def uppercase(cls, val):  # pylint: disable=no-self-use,no-self-argument
        """Turns to upper case"""
        if val:
            return val.upper()
        return val

    def get_client(self) -> str:
        """Concatenates client information"""
        return f"{self.client[0]}:{self.client[1]}"



"""Enums about HTTP"""

from taktile_types.enums.common import ExtendedEnum


class Method(ExtendedEnum):
    """HTTP Methods"""

    GET = "GET"
    HEAD = "HEAD"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    CONNECT = "CONNECT"
    PATCH = "PATCH"
    TRACE = "TRACE"
    OPTIONS = "OPTIONS"
